@@ -1,6 +1,6 @@
 /* enchant
  * Copyright (C) 2003-2004 Joan Moratinos <jmo@softcatala.org>, Dom Lachowicz
- * Copyright (C) 2016-2024 Reuben Thomas
+ * Copyright (C) 2016-2025 Reuben Thomas
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public
@@ -43,9 +43,14 @@
 
 #include "enchant-provider.h"
 
-#include <hunspell/hunspell.hxx>
+#pragma GCC diagnostic push
+#pragma GCC diagnostic ignored "-Weffc++"
+#include <hunspell.hxx>
+#pragma GCC diagnostic pop
+
 // hunspell itself uses this definition (which only supports the BMP)
 #define MAXWORDUTF8LEN (MAXWORDLEN * 3)
+#define DIC_SUFFIX ".dic"
 
 #include <glib.h>
 
@@ -59,6 +64,8 @@ static char *do_iconv(GIConv conv, const char *word) {
 	size_t len_in = strlen(in);
 	size_t len_out = len_in * 3;
 	char *out_buf = g_new0(char, len_out + 1);
+	if (out_buf == NULL)
+		return NULL;
 	char *out = out_buf;
 	size_t result = g_iconv(conv, &in, &len_in, &out, &len_out);
 	if (static_cast<size_t>(-1) == result)
@@ -154,18 +161,18 @@ HunspellChecker::suggestWord(const char* const utf8Word, size_t len, size_t *nsu
 
 	std::vector<std::string> sugMS = hunspell->suggest(out);
 	g_free(out);
-	*nsug = sugMS.size();
-	if (*nsug > 0) {
-		char **sug = g_new0 (char *, *nsug + 1);
-		for (size_t i=0, j=0; i<*nsug; i++) {
+	char **sug = g_new0 (char *, *nsug + 1);
+	if (sug) {
+		for (size_t i = 0, j = 0; i < *nsug; i++) {
 			const char *in = sugMS[i].c_str();
 			out = do_iconv(m_translate_out, in);
 			if (out != NULL)
 				sug[j++] = out;
 		}
-		return sug;
-	}
-	return nullptr;
+		*nsug = sugMS.size();
+	} else
+		*nsug = 0;
+	return sug;
 }
 
 void
@@ -205,31 +212,11 @@ s_buildDictionaryDirs (EnchantProvider * me, std::vector<std::string> & dirs)
 	dirs.push_back (tmp);
 	g_free(tmp);
 
-	for (const gchar* const * iter = g_get_system_data_dirs (); *iter; iter++)
-		{
-			tmp = g_build_filename (*iter, me->identify (me), nullptr);
-			dirs.push_back (tmp);
-			g_free(tmp);
-		}
-}
-
-static void
-s_buildHashNames (EnchantProvider * me, std::vector<std::string> & names, const char * dict)
-{
-	names.clear ();
-
-	std::vector<std::string> dirs;
-	s_buildDictionaryDirs (me, dirs);
-
-	char *dict_dic = g_strconcat(dict, ".dic", nullptr);
-	for (size_t i = 0; i < dirs.size(); i++)
-		{
-			char *tmp = g_build_filename (dirs[i].c_str(), dict_dic, nullptr);
-			names.push_back (tmp);
-			g_free (tmp);
-		}
-
-	g_free(dict_dic);
+	for (const gchar* const * iter = g_get_system_data_dirs (); *iter; iter++) {
+		tmp = g_build_filename (*iter, me->identify (me), nullptr);
+		dirs.push_back (tmp);
+		g_free(tmp);
+	}
 }
 
 static const std::string
@@ -246,94 +233,41 @@ s_fileExists(const std::string & file)
 	return g_file_test(file.c_str(), G_FILE_TEST_EXISTS) != 0;
 }
 
-static bool is_plausible_dict_for_tag(const char *dir_entry, const char *tag)
-{
-    const char *dic_suffix = ".dic";
-    size_t dic_suffix_len = strlen(dic_suffix);
-    size_t dir_entry_len = strlen(dir_entry);
-    size_t tag_len = strlen(tag);
-
-    if (dir_entry_len - dic_suffix_len < tag_len)
-	return false;
-    if (strcmp(dir_entry+dir_entry_len-dic_suffix_len, dic_suffix) != 0)
-	return false;
-    if (strncmp (dir_entry, tag, tag_len) != 0)
-	return false;
-    //e.g. requested dict for "fi",
-    //reject "fil_PH.dic"
-    //allow "fi-FOO.dic", "fi_FOO.dic", "fi.dic", etc.
-    if (!ispunct(dir_entry[tag_len]))
-	return false;
-    return true;
-}
-
 static char *
-hunspell_request_dictionary (EnchantProvider * me, const char * tag)
+hunspell_find_dictionary (EnchantProvider * me, const char * tag)
 {
-	std::vector<std::string> names;
-
-	s_buildHashNames (me, names, tag);
-
-	for (size_t i = 0; i < names.size (); i++) {
-		if (s_fileExists(names[i]) && s_fileExists(s_correspondingAffFile(names[i]))) {
-			return strdup (names[i].c_str());
-		}
-	}
-
 	std::vector<std::string> dirs;
 	s_buildDictionaryDirs (me, dirs);
 
+	std::string dict_basename(tag);
+	dict_basename += DIC_SUFFIX;
 	for (size_t i = 0; i < dirs.size(); i++) {
-		GError *err = NULL;
-		GDir *dir = g_dir_open (dirs[i].c_str(), 0, &err);
-		g_assert ((dir == NULL && err != NULL) || (dir != NULL && err == NULL));
-		if (err == NULL) {
-			const char *dir_entry;
-			while ((dir_entry = g_dir_read_name (dir)) != NULL) {
-				if (is_plausible_dict_for_tag(dir_entry, tag)) {
-					char *dict = g_build_filename (dirs[i].c_str(),
-								       dir_entry, nullptr);
-					if(s_fileExists(s_correspondingAffFile(dict))) {
-						g_dir_close (dir);
-						return dict;
-					} else {
-						g_debug ("hunspell provider: dictionary file %s has no corresponding affix file", dict);
-					}
-					g_free(dict);
-				}
-			}
+		char *filename = g_build_filename(dirs[i].c_str(), dict_basename.c_str(), nullptr);
+		if (s_fileExists(filename) && s_fileExists(s_correspondingAffFile(filename)))
+			return filename;
+                g_free(filename);
+        }
 
-			g_dir_close (dir);
-		} else {
-			g_debug ("hunspell provider: could not open directory %s: %s", dirs[i].c_str(), err->message);
-			g_error_free (err);
-		}
-	}
-
-	return NULL;
+	return nullptr;
 }
 
 bool
 HunspellChecker::requestDictionary(const char *szLang)
 {
-	char *dic = hunspell_request_dictionary (me, szLang);
+	char *dic = hunspell_find_dictionary (me, szLang);
 	if (!dic)
 		return false;
 
+        if (hunspell) {
+                delete hunspell;
+                free(wordchars);
+                wordchars = NULL;
+        }
 	std::string aff(s_correspondingAffFile(dic));
-	if (s_fileExists(aff))
-	{
-		if (hunspell) {
-			delete hunspell;
-			free(wordchars);
-			wordchars = NULL;
-		}
-		hunspell = new Hunspell(aff.c_str(), dic);
-	}
+        hunspell = new Hunspell(aff.c_str(), dic);
 	free(dic);
-	if(hunspell == NULL){
+	if(hunspell == NULL)
 		return false;
-	}
 	const char *enc = hunspell->get_dic_encoding();
 
 	m_translate_in = g_iconv_open(enc, "UTF-8");
@@ -402,9 +336,8 @@ hunspell_dict_is_word_character (EnchantDict *me, uint32_t uc, size_t n)
 	HunspellChecker * checker = static_cast<HunspellChecker *>(me->user_data);
 	/* Accept quote marks anywhere except at the end of a word, as per
 	   hunspell's textparser.cxx/TextParser::next_token */
-	if ((uc == g_utf8_get_char("'") || uc == g_utf8_get_char("’")) && checker->apostropheIsWordChar) {
+	if ((uc == g_utf8_get_char("'") || uc == g_utf8_get_char("’")) && checker->apostropheIsWordChar)
 		return n < 2;
-	}
 	return g_unichar_isalpha(uc) || g_utf8_strchr(checker->getWordchars(), -1, uc);
 }
 
@@ -418,20 +351,17 @@ hunspell_provider_enum_dicts (const char * const directory,
 		while ((entry = g_dir_read_name (dir)) != NULL) {
 			char * utf8_entry = g_filename_to_utf8 (entry, -1, nullptr, nullptr, nullptr);
 			if (utf8_entry) {
-				std::string dir_entry (utf8_entry);
-				g_free (utf8_entry);
+                                std::string dir_entry (utf8_entry);
+                                g_free (utf8_entry);
 
-				int hit = dir_entry.rfind (".dic");
-				if (hit != -1) {
-					/* don't include hyphenation dictionaries
-					   and require .aff file to be present*/
-					if(dir_entry.compare (0, 5, "hyph_") != 0)
-					{
-						char * dic = g_build_filename(directory, dir_entry.c_str(), nullptr);
-						if (s_fileExists(s_correspondingAffFile(dic)))
-						{
-							out_dicts.push_back (dir_entry.substr (0, hit));
-						}
+				std::string::size_type hit = dir_entry.rfind(DIC_SUFFIX);
+                                if (hit != std::string::npos) {
+                                        /* don't include hyphenation dictionaries
+                                           and require .aff file to be present*/
+                                        if (dir_entry.compare (0, 5, "hyph_") != 0) {
+                                                char * dic = g_build_filename(directory, dir_entry.c_str(), nullptr);
+						if (dic && s_fileExists(s_correspondingAffFile(dic)))
+                                                        out_dicts.push_back (dir_entry.substr (0, hit));
 						g_free(dic);
 					}
 				}
@@ -447,24 +377,20 @@ extern "C" {
 static char **
 hunspell_provider_list_dicts (EnchantProvider * me, size_t * out_n_dicts)
 {
+	// Enumerate dictionaries in each directory.
 	std::vector<std::string> dict_dirs, dicts;
-	char ** dictionary_list = NULL;
-
 	s_buildDictionaryDirs (me, dict_dirs);
-
 	for (size_t i = 0; i < dict_dirs.size(); i++)
-		{
-			hunspell_provider_enum_dicts (dict_dirs[i].c_str(), dicts);
-		}
-
-	if (dicts.size () > 0) {
-		dictionary_list = g_new0 (char *, dicts.size() + 1);
+		hunspell_provider_enum_dicts (dict_dirs[i].c_str(), dicts);
 
+	// Convert vector to array of pointers and return.
+	char ** dictionary_list = g_new0 (char *, dicts.size() + 1);
+	if (dictionary_list) {
 		for (size_t i = 0; i < dicts.size(); i++)
 			dictionary_list[i] = g_strdup (dicts[i].c_str());
-	}
-
-	*out_n_dicts = dicts.size ();
+		*out_n_dicts = dicts.size ();
+	} else
+		*out_n_dicts = 0;
 	return dictionary_list;
 }
 
@@ -481,7 +407,9 @@ hunspell_provider_request_dict(EnchantProvider * me, const char *const tag)
 		return NULL;
 	}
 
-	EnchantDict *dict = enchant_broker_new_dict(me->owner);
+	EnchantDict *dict = enchant_dict_new();
+	if (dict == NULL)
+		return NULL;
 	dict->user_data = (void *) checker;
 	dict->check = hunspell_dict_check;
 	dict->suggest = hunspell_dict_suggest;
@@ -504,16 +432,7 @@ static int
 hunspell_provider_dictionary_exists (EnchantProvider * me,
 				     const char *const tag)
 {
-	std::vector <std::string> names;
-	s_buildHashNames (me, names, tag);
-	for (size_t i = 0; i < names.size(); i++) {
-		if (s_fileExists(names[i]) && s_fileExists(s_correspondingAffFile(names[i])))
-		{
-			return 1;
-		}
-	}
-
-	return 0;
+	return hunspell_find_dictionary(me, tag) != NULL;
 }
 
 static const char *
